@@ -1,10 +1,12 @@
 use crate::core::Target;
 use crate::util::errors::{CargoResult, CargoResultExt};
 use crate::util::interning::InternedString;
-use crate::util::Config;
+use crate::util::{Config, StableHasher};
 use anyhow::bail;
 use serde::Serialize;
 use std::collections::BTreeSet;
+use std::fs;
+use std::hash::{Hash, Hasher};
 use std::path::Path;
 
 /// Indicator for how a unit is being compiled.
@@ -27,10 +29,7 @@ pub enum CompileKind {
 
 impl CompileKind {
     pub fn is_host(&self) -> bool {
-        match self {
-            CompileKind::Host => true,
-            _ => false,
-        }
+        matches!(self, CompileKind::Host)
     }
 
     pub fn for_target(self, target: &Target) -> CompileKind {
@@ -81,6 +80,18 @@ impl CompileKind {
         };
         Ok(vec![kind])
     }
+
+    /// Hash used for fingerprinting.
+    ///
+    /// Metadata hashing uses the normal Hash trait, which does not
+    /// differentiate on `.json` file contents. The fingerprint hash does
+    /// check the contents.
+    pub fn fingerprint_hash(&self) -> u64 {
+        match self {
+            CompileKind::Host => 0,
+            CompileKind::Target(target) => target.fingerprint_hash(),
+        }
+    }
 }
 
 impl serde::ser::Serialize for CompileKind {
@@ -169,4 +180,19 @@ impl CompileTarget {
             &self.name
         }
     }
+
+    /// See [`CompileKind::fingerprint_hash`].
+    pub fn fingerprint_hash(&self) -> u64 {
+        let mut hasher = StableHasher::new();
+        self.name.hash(&mut hasher);
+        if self.name.ends_with(".json") {
+            // This may have some performance concerns, since it is called
+            // fairly often. If that ever seems worth fixing, consider
+            // embedding this in `CompileTarget`.
+            if let Ok(contents) = fs::read_to_string(self.name) {
+                contents.hash(&mut hasher);
+            }
+        }
+        hasher.finish()
+    }
 }
