@@ -4,7 +4,7 @@ pub use self::canonical_url::CanonicalUrl;
 pub use self::config::{homedir, Config, ConfigValue};
 pub use self::dependency_queue::DependencyQueue;
 pub use self::diagnostic_server::RustfixDiagnosticServer;
-pub use self::errors::{internal, process_error};
+pub use self::errors::{exit_status_to_string, internal, process_error, process_error_raw};
 pub use self::errors::{CargoResult, CargoResultExt, CliResult, Test};
 pub use self::errors::{CargoTestError, CliError, ProcessError};
 pub use self::flock::{FileLock, Filesystem};
@@ -27,8 +27,8 @@ pub use self::sha256::Sha256;
 pub use self::to_semver::ToSemver;
 pub use self::vcs::{existing_vcs_repo, FossilRepo, GitRepo, HgRepo, PijulRepo};
 pub use self::workspace::{
-    print_available_benches, print_available_binaries, print_available_examples,
-    print_available_packages, print_available_tests,
+    add_path_args, path_args, print_available_benches, print_available_binaries,
+    print_available_examples, print_available_packages, print_available_tests,
 };
 
 mod canonical_url;
@@ -79,3 +79,15 @@ pub fn elapsed(duration: Duration) -> String {
 pub fn is_ci() -> bool {
     std::env::var("CI").is_ok() || std::env::var("TF_BUILD").is_ok()
 }
+
+pub fn indented_lines(text: &str) -> String {
+    text.lines()
+        .map(|line| {
+            if line.is_empty() {
+                String::from("\n")
+            } else {
+                format!("  {}\n", line)
+            }
+        })
+        .collect()
+}
