@@ -24,6 +24,7 @@ mod build_script_extra_link_arg;
 mod cache_messages;
 mod cargo_alias_config;
 mod cargo_command;
+mod cargo_env_config;
 mod cargo_features;
 mod cargo_targets;
 mod cfg;
@@ -51,6 +52,7 @@ mod features_namespaced;
 mod fetch;
 mod fix;
 mod freshness;
+mod future_incompat_report;
 mod generate_lockfile;
 mod git;
 mod git_auth;
@@ -78,6 +80,7 @@ mod multitarget;
 mod net_config;
 mod new;
 mod offline;
+mod old_cargos;
 mod out_dir;
 mod owner;
 mod package;
