@@ -1,7 +1,6 @@
 //! Tests for configuration values that point to programs.
 
-use cargo_test_support::rustc_host;
-use cargo_test_support::{basic_lib_manifest, project};
+use cargo_test_support::{basic_lib_manifest, no_such_file_err_msg, project, rustc_host};
 
 #[cargo_test]
 fn pathless_tools() {
@@ -14,10 +13,9 @@ fn pathless_tools() {
             ".cargo/config",
             &format!(
                 r#"
-            [target.{}]
-            ar = "nonexistent-ar"
-            linker = "nonexistent-linker"
-        "#,
+                    [target.{}]
+                    linker = "nonexistent-linker"
+                "#,
                 target
             ),
         )
@@ -27,7 +25,7 @@ fn pathless_tools() {
         .with_stderr(
             "\
 [COMPILING] foo v0.5.0 ([CWD])
-[RUNNING] `rustc [..] -C ar=nonexistent-ar -C linker=nonexistent-linker [..]`
+[RUNNING] `rustc [..] -C linker=nonexistent-linker [..]`
 [FINISHED] dev [unoptimized + debuginfo] target(s) in [..]
 ",
         )
@@ -39,13 +37,10 @@ fn absolute_tools() {
     let target = rustc_host();
 
     // Escaped as they appear within a TOML config file
-    let config = if cfg!(windows) {
-        (
-            r#"C:\\bogus\\nonexistent-ar"#,
-            r#"C:\\bogus\\nonexistent-linker"#,
-        )
+    let linker = if cfg!(windows) {
+        r#"C:\\bogus\\nonexistent-linker"#
     } else {
-        (r#"/bogus/nonexistent-ar"#, r#"/bogus/nonexistent-linker"#)
+        r#"/bogus/nonexistent-linker"#
     };
 
     let foo = project()
@@ -55,13 +50,11 @@ fn absolute_tools() {
             ".cargo/config",
             &format!(
                 r#"
-            [target.{target}]
-            ar = "{ar}"
-            linker = "{linker}"
-        "#,
+                    [target.{target}]
+                    linker = "{linker}"
+                "#,
                 target = target,
-                ar = config.0,
-                linker = config.1
+                linker = linker
             ),
         )
         .build();
@@ -70,7 +63,7 @@ fn absolute_tools() {
         .with_stderr(
             "\
 [COMPILING] foo v0.5.0 ([CWD])
-[RUNNING] `rustc [..] -C ar=[..]bogus/nonexistent-ar -C linker=[..]bogus/nonexistent-linker [..]`
+[RUNNING] `rustc [..] -C linker=[..]bogus/nonexistent-linker [..]`
 [FINISHED] dev [unoptimized + debuginfo] target(s) in [..]
 ",
         )
@@ -82,10 +75,10 @@ fn relative_tools() {
     let target = rustc_host();
 
     // Escaped as they appear within a TOML config file
-    let config = if cfg!(windows) {
-        (r#".\\nonexistent-ar"#, r#".\\tools\\nonexistent-linker"#)
+    let linker = if cfg!(windows) {
+        r#".\\tools\\nonexistent-linker"#
     } else {
-        (r#"./nonexistent-ar"#, r#"./tools/nonexistent-linker"#)
+        r#"./tools/nonexistent-linker"#
     };
 
     // Funky directory structure to test that relative tool paths are made absolute
@@ -98,27 +91,28 @@ fn relative_tools() {
             ".cargo/config",
             &format!(
                 r#"
-            [target.{target}]
-            ar = "{ar}"
-            linker = "{linker}"
-        "#,
+                    [target.{target}]
+                    linker = "{linker}"
+                "#,
                 target = target,
-                ar = config.0,
-                linker = config.1
+                linker = linker
             ),
         )
         .build();
 
     let prefix = p.root().into_os_string().into_string().unwrap();
 
-    p.cargo("build --verbose").cwd("bar").with_stderr(&format!(
+    p.cargo("build --verbose")
+        .cwd("bar")
+        .with_stderr(&format!(
             "\
 [COMPILING] bar v0.5.0 ([CWD])
-[RUNNING] `rustc [..] -C ar={prefix}/./nonexistent-ar -C linker={prefix}/./tools/nonexistent-linker [..]`
+[RUNNING] `rustc [..] -C linker={prefix}/./tools/nonexistent-linker [..]`
 [FINISHED] dev [unoptimized + debuginfo] target(s) in [..]
 ",
             prefix = prefix,
-        )).run();
+        ))
+        .run();
 }
 
 #[cargo_test]
@@ -133,9 +127,9 @@ fn custom_runner() {
             ".cargo/config",
             &format!(
                 r#"
-            [target.{}]
-            runner = "nonexistent-runner -r"
-        "#,
+                    [target.{}]
+                    runner = "nonexistent-runner -r"
+                "#,
                 target
             ),
         )
@@ -204,7 +198,7 @@ fn custom_runner_cfg() {
         .run();
 }
 
-// custom runner set via `target.$triple.runner` have precende over `target.'cfg(..)'.runner`
+// custom runner set via `target.$triple.runner` have precedence over `target.'cfg(..)'.runner`
 #[cargo_test]
 fn custom_runner_cfg_precedence() {
     let target = rustc_host();
@@ -215,12 +209,12 @@ fn custom_runner_cfg_precedence() {
             ".cargo/config",
             &format!(
                 r#"
-            [target.'cfg(not(target_os = "none"))']
-            runner = "ignored-runner"
+                    [target.'cfg(not(target_os = "none"))']
+                    runner = "ignored-runner"
 
-            [target.{}]
-            runner = "nonexistent-runner -r"
-        "#,
+                    [target.{}]
+                    runner = "nonexistent-runner -r"
+                "#,
                 target
             ),
         )
@@ -256,9 +250,171 @@ fn custom_runner_cfg_collision() {
 
     p.cargo("run -- --param")
         .with_status(101)
-        .with_stderr_contains(
+        .with_stderr(
             "\
 [ERROR] several matching instances of `target.'cfg(..)'.runner` in `.cargo/config`
+first match `cfg(not(target_arch = \"avr\"))` located in [..]/foo/.cargo/config
+second match `cfg(not(target_os = \"none\"))` located in [..]/foo/.cargo/config
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn custom_runner_env() {
+    let target = rustc_host();
+    let p = project().file("src/main.rs", "fn main() {}").build();
+
+    let key = format!(
+        "CARGO_TARGET_{}_RUNNER",
+        target.to_uppercase().replace('-', "_")
+    );
+
+    p.cargo("run")
+        .env(&key, "nonexistent-runner --foo")
+        .with_status(101)
+        .with_stderr(&format!(
+            "\
+[COMPILING] foo [..]
+[FINISHED] dev [..]
+[RUNNING] `nonexistent-runner --foo target/debug/foo[EXE]`
+[ERROR] could not execute process `nonexistent-runner --foo target/debug/foo[EXE]` (never executed)
+
+Caused by:
+  {}
+",
+            no_such_file_err_msg()
+        ))
+        .run();
+}
+
+#[cargo_test]
+fn custom_runner_env_overrides_config() {
+    let target = rustc_host();
+    let p = project()
+        .file("src/main.rs", "fn main() {}")
+        .file(
+            ".cargo/config.toml",
+            &format!(
+                r#"
+                    [target.{}]
+                    runner = "should-not-run -r"
+                "#,
+                target
+            ),
+        )
+        .build();
+
+    let key = format!(
+        "CARGO_TARGET_{}_RUNNER",
+        target.to_uppercase().replace('-', "_")
+    );
+
+    p.cargo("run")
+        .env(&key, "should-run --foo")
+        .with_status(101)
+        .with_stderr_contains("[RUNNING] `should-run --foo target/debug/foo[EXE]`")
+        .run();
+}
+
+#[cargo_test]
+#[cfg(unix)] // Assumes `true` is in PATH.
+fn custom_runner_env_true() {
+    // Check for a bug where "true" was interpreted as a boolean instead of
+    // the executable.
+    let target = rustc_host();
+    let p = project().file("src/main.rs", "fn main() {}").build();
+
+    let key = format!(
+        "CARGO_TARGET_{}_RUNNER",
+        target.to_uppercase().replace('-', "_")
+    );
+
+    p.cargo("run")
+        .env(&key, "true")
+        .with_stderr_contains("[RUNNING] `true target/debug/foo[EXE]`")
+        .run();
+}
+
+#[cargo_test]
+fn custom_linker_env() {
+    let target = rustc_host();
+    let p = project().file("src/main.rs", "fn main() {}").build();
+
+    let key = format!(
+        "CARGO_TARGET_{}_LINKER",
+        target.to_uppercase().replace('-', "_")
+    );
+
+    p.cargo("build -v")
+        .env(&key, "nonexistent-linker")
+        .with_status(101)
+        .with_stderr_contains("[RUNNING] `rustc [..]-C linker=nonexistent-linker [..]")
+        .run();
+}
+
+#[cargo_test]
+fn target_in_environment_contains_lower_case() {
+    let p = project().file("src/main.rs", "fn main() {}").build();
+
+    let target = rustc_host();
+    let env_key = format!(
+        "CARGO_TARGET_{}_LINKER",
+        target.to_lowercase().replace('-', "_")
+    );
+
+    let mut execs = p.cargo("build -v --target");
+    execs.arg(target).env(&env_key, "nonexistent-linker");
+    if cfg!(windows) {
+        // Windows env keys are case insensitive, so no warning, but it will
+        // fail due to the missing linker.
+        execs
+            .with_stderr_does_not_contain("warning:[..]")
+            .with_status(101);
+    } else {
+        execs.with_stderr_contains(format!(
+            "warning: Environment variables are expected to use uppercase letters and underscores, \
+            the variable `{}` will be ignored and have no effect",
+            env_key
+        ));
+    }
+    execs.run();
+}
+
+#[cargo_test]
+fn cfg_ignored_fields() {
+    // Test for some ignored fields in [target.'cfg()'] tables.
+    let p = project()
+        .file(
+            ".cargo/config",
+            r#"
+            # Try some empty tables.
+            [target.'cfg(not(foo))']
+            [target.'cfg(not(bar))'.somelib]
+
+            # A bunch of unused fields.
+            [target.'cfg(not(target_os = "none"))']
+            linker = 'false'
+            ar = 'false'
+            foo = {rustc-flags = "-l foo"}
+            invalid = 1
+            runner = 'false'
+            rustflags = ''
+            "#,
+        )
+        .file("src/lib.rs", "")
+        .build();
+
+    p.cargo("check")
+        .with_stderr(
+            "\
+[WARNING] unused key `somelib` in [target] config table `cfg(not(bar))`
+[WARNING] unused key `ar` in [target] config table `cfg(not(target_os = \"none\"))`
+[WARNING] unused key `foo` in [target] config table `cfg(not(target_os = \"none\"))`
+[WARNING] unused key `invalid` in [target] config table `cfg(not(target_os = \"none\"))`
+[WARNING] unused key `linker` in [target] config table `cfg(not(target_os = \"none\"))`
+[CHECKING] foo v0.0.1 ([..])
+[FINISHED] [..]
 ",
         )
         .run();
