@@ -430,41 +430,6 @@ fn update_precise_first_run() {
         .with_json(
             r#"{
   "packages": [
-    {
-      "authors": [],
-      "categories": [],
-      "dependencies": [],
-      "description": null,
-      "edition": "2015",
-      "features": {},
-      "id": "serde 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)",
-      "keywords": [],
-      "license": null,
-      "license_file": null,
-      "links": null,
-      "manifest_path": "[..]/home/.cargo/registry/src/-[..]/serde-0.2.0/Cargo.toml",
-      "metadata": null,
-      "publish": null,
-      "name": "serde",
-      "readme": null,
-      "repository": null,
-      "source": "registry+https://github.com/rust-lang/crates.io-index",
-      "targets": [
-        {
-          "crate_types": [
-            "lib"
-          ],
-          "doctest": true,
-          "edition": "2015",
-          "kind": [
-            "lib"
-          ],
-          "name": "serde",
-          "src_path": "[..]/home/.cargo/registry/src/-[..]/serde-0.2.0/src/lib.rs"
-        }
-      ],
-      "version": "0.2.0"
-    },
     {
       "authors": [],
       "categories": [],
@@ -483,8 +448,10 @@ fn update_precise_first_run() {
         }
       ],
       "description": null,
+      "documentation": null,
       "edition": "2015",
       "features": {},
+      "homepage": null,
       "id": "bar 0.0.1 (path+file://[..]/foo)",
       "keywords": [],
       "license": null,
@@ -502,7 +469,9 @@ fn update_precise_first_run() {
           "crate_types": [
             "lib"
           ],
+          "doc": true,
           "doctest": true,
+          "test": true,
           "edition": "2015",
           "kind": [
             "lib"
@@ -512,6 +481,45 @@ fn update_precise_first_run() {
         }
       ],
       "version": "0.0.1"
+    },
+    {
+      "authors": [],
+      "categories": [],
+      "dependencies": [],
+      "description": null,
+      "documentation": null,
+      "edition": "2015",
+      "features": {},
+      "homepage": null,
+      "id": "serde 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)",
+      "keywords": [],
+      "license": null,
+      "license_file": null,
+      "links": null,
+      "manifest_path": "[..]/home/.cargo/registry/src/-[..]/serde-0.2.0/Cargo.toml",
+      "metadata": null,
+      "publish": null,
+      "name": "serde",
+      "readme": null,
+      "repository": null,
+      "source": "registry+https://github.com/rust-lang/crates.io-index",
+      "targets": [
+        {
+          "crate_types": [
+            "lib"
+          ],
+          "doc": true,
+          "doctest": true,
+          "edition": "2015",
+          "kind": [
+            "lib"
+          ],
+          "name": "serde",
+          "src_path": "[..]/home/.cargo/registry/src/-[..]/serde-0.2.0/src/lib.rs",
+          "test": true
+        }
+      ],
+      "version": "0.2.0"
     }
   ],
   "resolve": {
@@ -577,6 +585,7 @@ fn preserve_top_comment() {
     lines.insert(2, "# some other comment");
     let mut lockfile = lines.join("\n");
     lockfile.push_str("\n\n"); // .lines/.join loses the last newline
+                               // >>>>>>> parent of 7dd9872c1... Change git dependencies to use `HEAD` by default
     println!("saving Cargo.lock contents:\n{}", lockfile);
 
     p.change_file("Cargo.lock", &lockfile);
@@ -643,3 +652,28 @@ fn dry_run_update() {
     let new_lockfile = p.read_lockfile();
     assert_eq!(old_lockfile, new_lockfile)
 }
+
+#[cargo_test]
+fn workspace_only() {
+    let p = project().file("src/main.rs", "fn main() {}").build();
+    p.cargo("generate-lockfile").run();
+    let lock1 = p.read_lockfile();
+
+    p.change_file(
+        "Cargo.toml",
+        r#"
+            [package]
+            name = "foo"
+            authors = []
+            version = "0.0.2"
+        "#,
+    );
+    p.cargo("update --workspace").run();
+    let lock2 = p.read_lockfile();
+
+    assert_ne!(lock1, lock2);
+    assert!(lock1.contains("0.0.1"));
+    assert!(lock2.contains("0.0.2"));
+    assert!(!lock1.contains("0.0.2"));
+    assert!(!lock2.contains("0.0.1"));
+}
